#include "BmpLoader.h"
#include <algorithm>
#include <cmath>

BmpInfo BmpLoader::readBmpInfo(const std::string& filePath) {
    std::ifstream file(filePath, std::ios::binary);
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open BMP file: " + filePath);
    }

    BmpHeader header;
    file.read(reinterpret_cast<char*>(&header), sizeof(header));

    if (header.signature != 0x4D42) { // 'BM'
        throw std::runtime_error("Invalid BMP signature in file: " + filePath);
    }

    BmpInfoHeader infoHeader;
    file.read(reinterpret_cast<char*>(&infoHeader), sizeof(infoHeader));

    BmpInfo info;
    info.width = infoHeader.width;
    info.height = std::abs(infoHeader.height); // Handle bottom-up or top-down BMPs
    info.bitsPerPixel = infoHeader.bitsPerPixel;
    info.dataOffset = header.dataOffset;

    // Calculate row padding (rows are padded to 4-byte boundary)
    int bytesPerPixel = info.bitsPerPixel / 8;
    info.rowSize = info.width * bytesPerPixel;
    info.rowPadding = (4 - (info.rowSize % 4)) % 4;

    file.close();

    return info;
}

std::vector<uint8_t> BmpLoader::readBmpRegion(
        const std::string& filePath,
        const BmpInfo& info,
        int32_t startX, int32_t startY,
        int32_t width, int32_t height) {

    std::ifstream file(filePath, std::ios::binary);
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open BMP file: " + filePath);
    }

    // Clamp region to image dimensions
    startX = std::max(0, startX);
    startY = std::max(0, startY);
    width = std::min(width, info.width - startX);
    height = std::min(height, info.height - startY);

    if (width <= 0 || height <= 0) {
        return std::vector<uint8_t>();
    }

    // Calculate bytes per pixel
    int bytesPerPixel = info.bitsPerPixel / 8;
    int regionRowSize = width * bytesPerPixel;

    // Allocate memory for the region
    std::vector<uint8_t> regionData(width * height * bytesPerPixel);

    // Read data row by row
    for (int y = 0; y < height; y++) {
        // Calculate source position in file
        int sourceY = (info.height - 1 - (startY + y)); // Account for BMP being stored bottom-up
        long offset = info.dataOffset + (sourceY * (info.rowSize + info.rowPadding)) + startX * bytesPerPixel;

        file.seekg(offset, std::ios::beg);
        file.read(reinterpret_cast<char*>(regionData.data() + y * regionRowSize), regionRowSize);
    }

    file.close();

    return regionData;
}

// Convert sRGB to linear
float srgbToLinear(float srgb) {
    if (srgb <= 0.04045f)
        return srgb / 12.92f;
    else
        return pow((srgb + 0.055f) / 1.055f, 2.4f);
}

std::vector<float> BmpLoader::readBmpRegionAsFloat(
        const std::string& filePath,
        const BmpInfo& info,
        int32_t startX, int32_t startY,
        int32_t width, int32_t height) {

    auto data = readBmpRegion(filePath, info, startX, startY, width, height);
    if (data.empty()) {
        return std::vector<float>();
    }

    int bytesPerPixel = info.bitsPerPixel / 8;
    std::vector<float> floatData(width * height * 3); // Always RGB float output

    // Convert each component from sRGB to linear space
#if 0
    for (int i = 0; i < width * height; i++) {
        floatData[i * 3 + 0] = srgbToLinear(data[i * bytesPerPixel + 2] / 255.0f); // R
        floatData[i * 3 + 1] = srgbToLinear(data[i * bytesPerPixel + 1] / 255.0f); // G
        floatData[i * 3 + 2] = srgbToLinear(data[i * bytesPerPixel + 0] / 255.0f); // B
    }
#endif

    for (int i = 0; i < width * height; i++) {
        if (bytesPerPixel == 3 || bytesPerPixel == 4) { // RGB or RGBA
            floatData[i * 3 + 0] = data[i * bytesPerPixel + 2] / 255.0f; // R (BGR format in BMP)
            floatData[i * 3 + 1] = data[i * bytesPerPixel + 1] / 255.0f; // G
            floatData[i * 3 + 2] = data[i * bytesPerPixel + 0] / 255.0f; // B
        } else if (bytesPerPixel == 1) { // Grayscale
            floatData[i * 3 + 0] = data[i] / 255.0f; // R
            floatData[i * 3 + 1] = data[i] / 255.0f; // G
            floatData[i * 3 + 2] = data[i] / 255.0f; // B
        }
    }

    return floatData;
}