//
// Created by Matthew on 2025/3/22.
//

#include "HdrWriter.h"

#include <cmath>
#include <algorithm>
#include <vector>

// Define STB_IMAGE_WRITE_IMPLEMENTATION in exactly one CPP file
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include "stb_image_write.h" // Download this header from https://github.com/nothings/stb

float toneMap(float value) {
    return value / (1.0f + value); // Reinhard tone mapping
}
float applyGammaCorrection(float value) {
    return pow(value, 1.0f / 2.2f); // Gamma correction
}

bool HdrWriter::writeRGBE(const std::string& filename,
                          const std::vector<float>& data,
                          int width, int height) {
    std::ofstream file(filename, std::ios::binary);
    if (!file.is_open()) {
        return false;
    }

    writeHeader(file, width, height);

    // Write pixel data in RGBE format
    for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
            int pixelIndex = (y * width + x) * 3;
            float r = data[pixelIndex];
            float g = data[pixelIndex + 1];
            float b = data[pixelIndex + 2];

            uint8_t rgbe[4];
            rgbeFromFloat(r, g, b, rgbe);
            file.write(reinterpret_cast<char*>(rgbe), 4);
        }
    }

    return true;
}

void HdrWriter::writeHeader(std::ofstream& file, int width, int height) {
    // Write Radiance HDR header
    file << "#?RADIANCE\n";
    file << "FORMAT=32-bit_rle_rgbe\n\n";
    file << "-Y " << height << " +X " << width << "\n";
}

void HdrWriter::rgbeFromFloat(float r, float g, float b, uint8_t rgbe[4]) {
    float v = std::max(r, std::max(g, b));

    if (v < 1e-32f) {
        rgbe[0] = rgbe[1] = rgbe[2] = rgbe[3] = 0;
    } else {
        int e;
        v = frexpf(v, &e) * 256.0f / v;
        rgbe[0] = static_cast<uint8_t>(r * v);
        rgbe[1] = static_cast<uint8_t>(g * v);
        rgbe[2] = static_cast<uint8_t>(b * v);
        rgbe[3] = static_cast<uint8_t>(e + 128);
    }
}


bool HdrWriter::writeRGB(const std::string& filename,
                         const std::vector<float>& data,
                         int width, int height,
                         Format format) {

    if (format == Format::BMP) {
        return writeBMP(filename, data, width, height);
    }

    // For PNG/JPEG via stb_image_write.h
    // (Add this implementation if needed)
    return false;
}

bool HdrWriter::writeBMP(const std::string& filename,
                         const std::vector<float>& data,
                         int width, int height) {
    std::ofstream file(filename, std::ios::binary);
    if (!file.is_open()) {
        return false;
    }

    // BMP requires rows to be padded to multiples of 4 bytes
    int paddingSize = (4 - (width * 3) % 4) % 4;
    int rowSize = width * 3 + paddingSize;
    int dataSize = rowSize * height;
    int fileSize = 54 + dataSize; // 54 bytes for header + pixel data

    // BMP File Header (14 bytes)
    uint8_t fileHeader[14] = {
            'B', 'M',                               // Signature
            (uint8_t)(fileSize), (uint8_t)(fileSize >> 8),     // File size in bytes
            (uint8_t)(fileSize >> 16), (uint8_t)(fileSize >> 24),
            0, 0, 0, 0,                             // Reserved
            54, 0, 0, 0                             // Offset to pixel data
    };

    // BMP Info Header (40 bytes)
    uint8_t infoHeader[40] = {
            40, 0, 0, 0,                            // Info header size
            (uint8_t)(width), (uint8_t)(width >> 8),           // Width
            (uint8_t)(width >> 16), (uint8_t)(width >> 24),
            (uint8_t)(height), (uint8_t)(height >> 8),         // Height (negative for top-down)
            (uint8_t)(height >> 16), (uint8_t)(height >> 24),
            1, 0,                                   // Number of color planes
            24, 0,                                  // Bits per pixel (24 for RGB)
            0, 0, 0, 0,                             // No compression
            0, 0, 0, 0,                             // Image size (can be 0 for no compression)
            0, 0, 0, 0,                             // X pixels per meter
            0, 0, 0, 0,                             // Y pixels per meter
            0, 0, 0, 0,                             // Total colors (default)
            0, 0, 0, 0                              // Important colors (default)
    };

    file.write(reinterpret_cast<char*>(fileHeader), 14);
    file.write(reinterpret_cast<char*>(infoHeader), 40);

    // Padding bytes (zeros)
    std::vector<uint8_t> padding(paddingSize, 0);

    // Write pixel data (BGR order, bottom-to-top for standard BMP)
    for (int y = height - 1; y >= 0; y--) { // BMP stores rows bottom-to-top
        for (int x = 0; x < width; x++) {
            int pixelIndex = (y * width + x) * 3;

            // Apply tone mapping and gamma correction
#if 0
            float r = toneMap(data[pixelIndex + 0]);
            float g = toneMap(data[pixelIndex + 1]);
            float b = toneMap(data[pixelIndex + 2]);
#endif

            float exposure = 1.5f; // Adjust this value as needed
            float r = toneMap(data[pixelIndex + 0] * exposure);
            float g = toneMap(data[pixelIndex + 1] * exposure);
            float b = toneMap(data[pixelIndex + 2] * exposure);

            r = applyGammaCorrection(r);
            g = applyGammaCorrection(g);
            b = applyGammaCorrection(b);

            // Clamp and convert to byte
            uint8_t pixelData[3] = {
                    static_cast<uint8_t>(std::min(1.0f, b) * 255.0f), // B
                    static_cast<uint8_t>(std::min(1.0f, g) * 255.0f), // G
                    static_cast<uint8_t>(std::min(1.0f, r) * 255.0f)  // R
            };

            file.write(reinterpret_cast<char*>(pixelData), 3);
        }

        // Write padding bytes
        if (paddingSize > 0) {
            file.write(reinterpret_cast<char*>(padding.data()), paddingSize);
        }
    }

    return file.good();
}
